import collections
import inspect
import json
import logging
import os
import sqlite3
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
)

import celery
import falcon
import falcon.request
import falcon.response
import jinja2

import brokkoly.database
import brokkoly.resource


__all__ = ['BrokkolyError', 'Brokkoly', 'producer']
__author__ = "Motoki Naruse"
__copyright__ = "Motoki Naruse"
__credits__ = ["Motoki Naruse"]
__email__ = "motoki@naru.se"
__license__ = "MIT"
__maintainer__ = "Motoki Naruse"
__version__ = "0.3.0"


Validation = List[Tuple[str, Any]]
Processor = collections.namedtuple('Processor', ['func', 'validation'])
Message = Dict[str, Any]

_tasks = collections.defaultdict(dict)  # type: collections.defaultdict


logger = logging.getLogger(__name__)


class BrokkolyError(Exception):
    pass


class Brokkoly:
    def __init__(self, name: str, broker: str) -> None:
        if name.startswith('_'):
            # Because the names is reserved for control.
            raise BrokkolyError("Queue name starting with _ is not allowed.")
        self.celery = celery.Celery(name, broker=broker)
        self._tasks = _tasks[name]

    def task(self, *preprocessors: Callable) -> Callable:
        def wrapper(f: Callable):
            if f.__name__ in self._tasks:
                raise BrokkolyError("{} is already registered".format(f.__name__))

            self._tasks[f.__name__] = (
                Processor(self.celery.task(f), _prepare_validation(f)),
                [
                    Processor(
                        preprocessor,
                        _prepare_validation(preprocessor)
                    )
                    for preprocessor in preprocessors
                ]
            )
            return f
        return wrapper


def _prepare_validation(f: Callable) -> Validation:
    fullspec = inspect.getfullargspec(f)

    args = {arg_name: Any for arg_name in fullspec.args}
    for arg_name, arg_type in fullspec.annotations.items():
        # returning type is doesn't matter. This is for input checking.
        if arg_name != 'return':
            args[arg_name] = arg_type

    return list(args.items())


def _validate(message: Message, validation: Validation) -> Message:
    validated = {}
    for arg_name, arg_type in validation:
        try:
            value = message[arg_name]
        except KeyError:
            raise falcon.HTTPBadRequest(
                "Missing required filed", "{} is required".format(arg_name))

        if arg_type != Any and not isinstance(value, arg_type):
            raise falcon.HTTPBadRequest(
                "Invalid type", "{} must be {} type".format(arg_name, arg_type.__name__))

        validated[arg_name] = value

    return validated


class HTMLRendler:
    def __init__(self) -> None:
        self._jinja2 = jinja2.Environment(loader=jinja2.ChoiceLoader([
            jinja2.PackageLoader(__name__, 'resources'),
            jinja2.FileSystemLoader('resources'),
        ]), extensions=[
            'jinja2_highlight.HighlightExtension'
        ])
        self._jinja2.filters['pretty_print_json'] = lambda source: json.dumps(
            json.loads(source), indent=4, sort_keys=True)

    def render(self, template: str, **kwargs) -> str:
        return self._jinja2.get_template(template).render(**kwargs)


class Producer:
    def __init__(self, rendler: HTMLRendler) -> None:
        self._rendler = rendler

    def _recurse(self, message: Message, preprocessors: List[Processor]) -> Message:
        if preprocessors:
            (preprocess, preprocess_validation), *tail = preprocessors
            return self._recurse(preprocess(**_validate(message, preprocess_validation)), tail)
        return message

    def _validate_queue_and_task(
            self, queue_name: str, task_name: str) -> Tuple[Processor, List[Processor]]:
        # _tasks is defaultdict, it deoesn't raise KeyError.
        if queue_name not in _tasks:
            raise falcon.HTTPBadRequest(
                "Undefined queue", "{} is undefined queue".format(queue_name))
        queue_tasks = _tasks[queue_name]

        try:
            return queue_tasks[task_name]
        except KeyError:
            raise falcon.HTTPBadRequest("Undefined task", "{} is undefined task".format(task_name))

    def on_post(
            self, req: falcon.request.Request, resp: falcon.response.Response, queue_name: str,
            task_name: str
    ) -> None:
        (task, validation), preprocessors = self._validate_queue_and_task(queue_name, task_name)
        payload = req.stream.read().decode()
        if not payload:
            raise falcon.HTTPBadRequest(
                "Empty payload",
                "Even your task doesn't need any arguments, payload must have message filed"
            )

        try:
            message = json.loads(payload)['message']
        except ValueError:  # Python 3.4 doesn't have json.JSONDecodeError
            raise falcon.HTTPBadRequest("Payload is not a JSON", "The payload must be a JSON")
        except KeyError:
            raise falcon.HTTPBadRequest("Invalid JSON", "JSON must have message field")

        task.apply_async(
            kwargs=_validate(self._recurse(message, preprocessors), validation),
            serializer='json',
            compression='zlib'
        )
        brokkoly.database.MessageLog.create(queue_name, task_name, json.dumps(message))
        brokkoly.database.MessageLog.eliminate(queue_name, task_name)
        resp.status = falcon.HTTP_202
        resp.body = "{}"

    def on_get(
            self, req: falcon.request.Request, resp: falcon.response.Response, queue_name: str,
            task_name: str
    ) -> None:
        self._validate_queue_and_task(queue_name, task_name)
        messages = brokkoly.database.MessageLog.list_by_queue_name_and_task_name(
            queue_name, task_name)

        resp.content_type = 'text/html'
        resp.body = self._rendler.render(
            "enqueue.html", queue_name=queue_name, task_name=task_name, messages=messages)


class TaskListResource:
    def __init__(self, rendler: HTMLRendler) -> None:
        self._rendler = rendler

    def on_get(
            self, req: falcon.request.Request, resp: falcon.response.Response, queue_name: str
    ) -> None:
        try:
            task_names = self._list_task_name(queue_name)
        except KeyError:
            raise falcon.HTTPNotFound(
                title="Undefined queue",
                description="{} is undefined queue".format(queue_name)
            )

        resp.content_type = 'text/html'
        resp.body = self._rendler.render(
            "task_list.html", queue_name=queue_name, task_names=task_names)

    def _list_task_name(self, queue_name: str) -> Iterable[str]:
        # Check with "in" because _tasks is defaultdict.
        if queue_name in _tasks:
            return _tasks[queue_name].keys()
        raise KeyError


class QueueListResource:
    def __init__(self, rendler: HTMLRendler) -> None:
        self._rendler = rendler

    def on_get(self, req: falcon.request.Request, resp: falcon.response.Response) -> None:
        resp.content_type = 'text/html'
        resp.body = self._rendler.render("queue_list.html", queue_names=self._list_queue_name())

    def _list_queue_name(self) -> List[str]:
        return sorted(_tasks.keys())


class StaticResource:
    def _read_resource(self, filename: str) -> bytes:
        with open(brokkoly.resource.resource_filename(filename), 'rb') as f:
            return f.read()

    def on_get(
            self, req: falcon.request.Request, resp: falcon.response.Response, filename: str
    ) -> None:
        resp.content_type = {
            '.css': "text/css",
            '.js': "application/javascript",
        }[os.path.splitext(filename)[1]]

        resp.body = self._read_resource(filename)


class DBManager:
    def __init__(
            self, connection_manager: brokkoly.database.ThreadLocalDBConnectionManager) -> None:
        self.connection_manager = connection_manager

    def process_resource(
            self, req: falcon.request.Request, resp: falcon.response.Response, resource, params
    ) -> None:
        self.connection_manager.reconnect()

    def process_response(
            self, req: falcon.request.Request, resp: falcon.response.Response, resource,
            req_succeeded: bool
    ) -> None:
        connection = self.connection_manager.get()
        if not connection:
            # make connection only requested URL is matched any route.
            return

        if req_succeeded:
            connection.commit()
            connection.close()
            return

        # I think no way to know the connection is alive or not. When the thread is reused, and
        # it was passed process_resource method, connection is not None and comes here.
        try:
            connection.rollback()
            connection.close()
        except sqlite3.ProgrammingError:
            # With above reason, I think we don't need to report this one as exception.
            logger.debug("Failed to rollback or close SQLite3 connection.")


def init_logger(log_level: int) -> None:
    logging.basicConfig(
        level=log_level,
        format="%(asctime)s %(levelname)-5.5s [%(name)s:%(lineno)s] %(message)s"
    )


def producer(*, path: Optional[str]=None, log_level=logging.ERROR) -> falcon.api.API:
    init_logger(log_level)
    brokkoly.database.db.dbname = "brokkoly.db"

    brokkoly.database.Migrator(__version__).migrate()

    application = falcon.API(middleware=[DBManager(brokkoly.database.db)])
    rendler = HTMLRendler()
    for controller, route in [
            (StaticResource(), "/__static__/{filename}"),
            (Producer(rendler), "/{queue_name}/{task_name}"),
            (QueueListResource(rendler), "/"),
            (TaskListResource(rendler), "/{queue_name}"),
    ]:
        application.add_route("/{}{}".format(path, route) if path else route, controller)

    return application



import contextlib
import datetime
import logging
import os
import sqlite3
import threading
from typing import (
    Iterator,
    Optional,
)

import brokkoly.resource


logger = logging.getLogger(__name__)


class ThreadLocalDBConnectionManager:
    _connections = {}  # type: Dict[int, sqlite3.Connection]
    dbname = None  # type: Optional[str]

    def get(self) -> Optional[sqlite3.Connection]:
        return self._connections.get(threading.get_ident())

    def reconnect(self) -> None:
        id = threading.get_ident()
        connection = sqlite3.connect(self.dbname)
        connection.row_factory = sqlite3.Row
        logger.debug("Connect sqlite3 (%s) for %s", connection, id)
        self._connections[id] = connection


db = ThreadLocalDBConnectionManager()


class Migrator:
    def __init__(self, brokkoly_version: str) -> None:
        self.brokkoly_version = brokkoly_version

    def _has_database(self):
        with contextlib.closing(db.get().cursor()) as cursor:
            cursor.execute("""
            SELECT EXISTS(
                SELECT * FROM sqlite_master
                WHERE
                    sqlite_master.type = 'table' AND
                    sqlite_master.name = 'migrations'
            );
            """)
            return cursor.fetchone()[0]

    def _get_migration_version(self):
        with contextlib.closing(db.get().cursor()) as cursor:
            cursor.execute("""
            SELECT version
            FROM migrations
            ORDER BY version DESC
            LIMIT 1
            """)
            return cursor.fetchone()[0]

    def _iter_diff(self, version: str) -> Iterator[str]:
        """Returning data must be sorted.
        """
        logger.info("resource_dir: %s", brokkoly.resource.resource_dir)
        migration_dir = os.path.join(brokkoly.resource.resource_dir, 'migrations')
        return (
            os.path.join(migration_dir, filename)
            for filename in sorted(os.listdir(migration_dir))
            if os.path.splitext(filename)[0] > version
        )

    def _raise_for_invalid_version(self, schema_version: str) -> None:
        if schema_version > self.brokkoly_version:
            raise brokkoly.BrokkolyError(
                "Blokkory version: {}, Database schema version: {}. The database is setup with  "
                "newer version of Brokkoly. It doesn't provide downgrade database. Please upgrade "
                "Brokkoly.".format(self.brokkoly_version, schema_version)
            )

    def _run_migration_sql_file(self, filename: str) -> None:
        with open(filename, 'r') as f:
            sql = f.read()

        with contextlib.closing(db.get().cursor()) as cursor:
            try:
                cursor.executescript(sql)
            except sqlite3.Error as e:
                logger.exception("Failed to run migration: %s", filename)
                raise brokkoly.BrokkolyError("Failed to run migration") from e

    def _migrate(self) -> None:
        schema_version = self._get_migration_version() if self._has_database() else '0'
        logger.info("schema_version: %s", schema_version)
        self._raise_for_invalid_version(schema_version)

        for sql_file in self._iter_diff(schema_version):
            logger.info("Run migration: %s", sql_file)
            self._run_migration_sql_file(sql_file)

    def migrate(self) -> None:
        db.reconnect()
        try:
            self._migrate()
        finally:
            db.get().close()


class MessageLog:
    def __init__(
            self, *, id: int=None, queue_name: str, task_name: str, message: str,
            created_at: datetime.datetime
    ) -> None:
        self.id = id
        self.queue_name = queue_name
        self.task_name = task_name
        self.message = message
        self.created_at = created_at

    @classmethod
    def get_by_id(cls, id: int) -> Optional['MessageLog']:
        with contextlib.closing(db.get().cursor()) as cursor:
            cursor.execute("SELECT * FROM message_logs WHERE message_logs.id = ?", (id, ))
            return cls.from_sqlite3_row(cursor.fetchone())

    @classmethod
    def create(cls, queue_name: str, task_name: str, message: str) -> 'MessageLog':
        with contextlib.closing(db.get().cursor()) as cursor:
            cursor.execute("""
            INSERT INTO message_logs (queue_name, task_name, message)
            VALUES (?, ?, ?)
            ;""", (queue_name, task_name, message, ))
            # If SQLite3 supports "returning", I can use it here.
            id = cursor.lastrowid

        return cls.get_by_id(id)

    @classmethod
    def list_by_queue_name_and_task_name(
            cls, queue_name: str, task_name: str) -> Iterator['MessageLog']:
        with contextlib.closing(db.get().cursor()) as cursor:
            cursor.execute("""
            SELECT *
            FROM message_logs
            WHERE
                message_logs.queue_name = ? AND
                message_logs.task_name = ?
            ORDER BY message_logs.created_at DESC
            ;""", (queue_name, task_name, ))

            return (cls.from_sqlite3_row(row) for row in cursor.fetchall())

    @classmethod
    def from_sqlite3_row(cls, row: Optional[sqlite3.Row]) -> Optional['MessageLog']:
        return cls(**row) if row else None  # type: ignore

    @classmethod
    def eliminate(cls, queue_name: str, task_name: str) -> None:
        with contextlib.closing(db.get().cursor()) as cursor:
            cursor.execute("""
            DELETE FROM message_logs
            WHERE
                queue_name = :queue_name AND
                task_name = :task_name AND
                id NOT IN (
                    SELECT id
                    FROM message_logs
                    WHERE
                        queue_name = :queue_name AND
                        task_name = :task_name
                    ORDER BY created_at DESC
                    LIMIT 1000
                )
            ;
            """, {
                'queue_name': queue_name,
                'task_name': task_name,
            })



#!/usr/bin/env python3
import os
import re
import setuptools
import sys


here = os.path.abspath(os.path.dirname(__file__))


# Python 3.4 doesn't have typing module. So I cannot give typehint here.
def get_meta():
    with open(os.path.join(here, 'brokkoly/__init__.py')) as f:
        source = f.read()

    regex = r'^{}\s*=\s*[\'"]([^\'"]*)[\'"]'
    return lambda name: re.search(regex.format(name), source, re.MULTILINE).group(1)


get_meta = get_meta()

with open(os.path.join(here, 'README.rst')) as f:
    readme = f.read()

install_requires = [
    'celery',
    'falcon',
    'jinja2',
    'jinja2-highlight',
]

if sys.version_info < (3, 5):
    install_requires.append('typing')

test_requires = [
    'Pygments',
    'docutils',
    'flake8',
    'mypy',
    'pytest',
    'pytest-cov',
]

setuptools.setup(
    name='brokkoly',
    version=get_meta('__version__'),
    description="Brokkoly is a framework for enqueuing messages via HTTP request for celery.",
    long_description=readme,
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Software Development :: Libraries"
    ],
    keywords=["celery", "queue"],
    author=get_meta('__author__'),
    author_email=get_meta('__email__'),
    url="https://github.com/narusemotoki/brokkoly",
    license=get_meta('__license__'),
    packages=['brokkoly'],
    install_requires=install_requires,
    extras_require={
        'test': test_requires,
    },
    include_package_data=True,
)
